/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Calendar Palm Sync code.
 *
 * The Initial Developer of the Original Code is
 * Michiel van Leeuwen (mvl@exedo.nl).
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function nsLocalDatebookSyncRecord(aId, aDeleted, aCategory, aLocalEvent, aLastSync) { 
dump("c id: "+aId+"\n");
  this.mId = aId;
  this.mDeleted = aDeleted;
  this.mLocalEvent = aLocalEvent;

  this.mDirty = false;
  try {
    if ((aLocalEvent.lastModified / 1000) > aLastSync)
      this.mDirty = true;
  } catch(e) {
    // Looks like the record wasn't modified at all. assume it is new
    this.mDirty = true
  }
dump(aLocalEvent.title+" = "+this.mDirty+"\n");
dump((aLocalEvent.lastModified / 1000) + " ?? " + aLastSync +"\n");
}

nsLocalDatebookSyncRecord.prototype = {

  get id()       { return this.mId; },
  get deleted()  { return this.mDeleted; },
  get dirty()    { return this.mDirty; },

  get category() {
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  },

  get localEvent() { return this.mLocalEvent; },

  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.nsILocalDatebookSyncRecord) &&
        !iid.equals(Components.interfaces.nsISyncRecord) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
  },
  
  mId: 0,
  mDeleted: false,
  mDirty: false,
  mLocalEvent: null
}


const nsIPalmConnection = Components.interfaces.nsIPalmConnection;

function nsPalmDatebookEnumerator(aPalmConnection) {
  this.mPalmConnection = aPalmConnection;
  this.prefetch();
}

nsPalmDatebookEnumerator.prototype = {

  hasMoreElements: function() {
    return (this.mNext != null);
  },

  getNext: function() {
    var next = this.mNext;
    this.mIndex++;
    this.prefetch();
    return next;
  },
  
  prefetch: function() {
    try {
      this.mNext = this.mPalmConnection.readRecordByIndex(this.mIndex);
      this.mNext = this.mNext.QueryInterface(Components.interfaces.nsIPalmDatebookSyncRecord);
    } catch(e) {
      this.mNext = null;
    }
  },
  
  mPalmConnection: null,
  mNext: null,
  mIndex: 0
}

function nsLocalDatebookEnumerator(aDatebookEnum, aLastSync) {
  this.mDatebookEnum = aDatebookEnum;
  this.mLastSync = aLastSync;
}

nsLocalDatebookEnumerator.prototype = {
  hasMoreElements: function() {
    return (this.mDatebookEnum.hasMoreElements());
  },

  getNext: function() {
    var nextLocalEvent = this.mDatebookEnum.getNext().QueryInterface(Components.interfaces.oeIICalEvent);
dump("next id: "+nextLocalEvent.id+"\n");
dump("next title: "+nextLocalEvent.title+"\n");
    return new nsLocalDatebookSyncRecord(nextLocalEvent.syncId,
                                         false, 0,
                                         nextLocalEvent, this.mLastSync);
  },
  
  mDatebookEnum: null,
  mLastSync: 0
}


function nsPalmDatebookSync() { }

nsPalmDatebookSync.prototype = {

  get identifier() { return "PalmDatebookSync"; },

  get fastPossible() { return this.mFastPossible; },

  init: function () {
    this.mPalmConnection = Components.classes["@mozilla.org/sync/palmconnection;1"]
                             .createInstance(Components.interfaces.nsIPalmConnection);
    this.mPalmConnection.connect();
    this.mPalmConnection.openDataBase(nsIPalmConnection.PALM_DATEBOOK);

    var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                  getService(Components.interfaces.nsIPrefService);
    var prefbranch = prefs.getBranch(null);

    try {
      this.mCalendarPath = prefbranch.getCharPref("calendar.sync.server");
    } catch(e) {}
    dump("path: "+this.mCalendarPath+"\n");

    var icalContainer = Components.classes["@mozilla.org/ical-container;1"]
                           .getService(Components.interfaces.oeIICalContainer);
    this.mCalendar = icalContainer.getCalendar(this.mCalendarPath);

    dump("userID:       "+this.mPalmConnection.userID+"\n");
    dump("viewerID:     "+this.mPalmConnection.viewerID+"\n");

    try {
      this.mLastSync = prefbranch.getIntPref("calendar.sync.lastsync.PalmDatebookSync");
    } catch(e) {
      this.mLastSync = 0;
    }

    var now = new Date();
    this.mLastSyncInUTC = this.mLastSync + now.getTimezoneOffset() * 60;

    try {
      pcId = prefbranch.getIntPref("calendar.sync.palm.pcid");
    } catch(e) {
      var pcId = 0;
    }

    this.mFastPossible = false;
    // XXX is only lastSyncPC engouh?
    if ((this.mLastSync == this.mPalmConnection.lastSyncDate) &&
        (pcId == this.mPalmConnection.lastSyncPC))
      this.mFastPossible = true;

    dump("localLastSync:"+this.mLastSync+"\n");
    dump("localLastSync:"+this.mLastSyncInUTC+"\n");
    dump("lastSync:     "+this.mPalmConnection.lastSyncDate+"\n");
    dump("pcID:         "+pcId+"\n");
    dump("lastSyncPC:   "+this.mPalmConnection.lastSyncPC+"\n");
    dump("FastPossible: "+this.mFastPossible+"\n");
  },

  finish: function() {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                  getService(Components.interfaces.nsIPrefService);
    var prefbranch = prefs.getBranch(null);
    now = new Date();
    prefbranch.setIntPref("calendar.sync.lastsync.PalmDatebookSync", now.getTime() / 1000);

    var pcId = 0;
    try {
      pcId = prefbranch.getIntPref("calendar.sync.palm.pcid");
    } catch(e) {}

    if (!pcId) {
      pcId = Math.round(Math.random()*1000000);
      prefbranch.setIntPref("calendar.sync.palm.pcid", pcId);
    }

    this.mPalmConnection.lastSyncPC = pcId;
    this.mPalmConnection.successfulSyncDate = now.getTime() / 1000;
    this.mPalmConnection.lastSyncDate = now.getTime() / 1000;
    this.mPalmConnection.writeUserInfo();

    this.mPalmConnection.closeDataBase();
    this.mPalmConnection.disconnect();
  },


  get localRecords() {
    return new nsLocalDatebookEnumerator(this.mCalendar.getAllEvents(), this.mLastSync);
  },

  get remoteRecords() {
    return new nsPalmDatebookEnumerator(this.mPalmConnection);
  },

  getLocalRecordBySyncId: function(aId) {
    var allEvents = this.mCalendar.getAllEvents();
    while (allEvents.hasMoreElements()) {
      var record = allEvents.getNext().QueryInterface(Components.interfaces.oeIICalEvent);
      if (record.syncId && (record.syncId == aId)) {
        return new nsLocalDatebookSyncRecord(record.syncId, false, 0,
                                             record, this.mLastSyncInUTC);
      }
    }
    return null;
  },

  getRemoteRecordBySyncId: function(aId) {
    return this.mPalmConnection.readRecordById(aId);
  },

  addLocal: function (aRecord) {
    remoteRecord = aRecord.QueryInterface(Components.interfaces.nsIPalmDatebookSyncRecord)
    remoteRecord.remoteEvent.syncId = remoteRecord.id;
dump("Add local: "+remoteRecord.id+"\n");
    // pretent the alarm has fired when importing an event.
    // Otherwise, the calendar looses its mind and crashes.
    if (remoteRecord.remoteEvent.alarm &&
        (remoteRecord.remoteEvent.start.getTime() < Date.now()))
      remoteRecord.remoteEvent.lastAlarmAck = remoteRecord.remoteEvent.start.getTime()
    remoteRecord.remoteEvent.updateLastModified();
    var idstr = this.mCalendar.addEvent(remoteRecord.remoteEvent);
    dump("Added "+idstr+"\n");
  },

  addRemote: function (aRecord) {
dump("addRemote "+aRecord.id+"\n");
    var localRecord = aRecord.QueryInterface(Components.interfaces.nsILocalDatebookSyncRecord);

    var remoteRecord = this.mPalmConnection.newRecord()
                         .QueryInterface(Components.interfaces.nsIPalmDatebookSyncRecord);
    this.copyEvent(localRecord.localEvent, remoteRecord.remoteEvent);
    // Don't set an ID when writing. the handheld will find one.
    // We don't know what id is still unique for the handheld.
    remoteRecord.id = 0;
    this.mPalmConnection.writeRecord(remoteRecord);
    localRecord.localEvent.syncId = remoteRecord.id;
    localRecord.localEvent.updateLastModified();
    this.mCalendar.modifyEvent(localRecord.localEvent);
dump("Tried to add remote "+localRecord.localEvent.description+" to id "+remoteRecord.id+"\n");
  },

  removeLocal: function (aRecord) {
    var localEvent = aRecord.QueryInterface(Components.interfaces.nsILocalDatebookSyncRecord).localEvent;
dump("trying to delete "+localEvent.id+"\n");
    try {
      // in a try block, maybe it already was deleted locally
      this.mCalendar.deleteEvent(localEvent.id);
    } catch(e) {}
  },

  removeRemote: function (aRecord) {
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  },

  updateLocal: function (aRemoteRecord, aLocalRecord) {
    var localRecord = aLocalRecord.QueryInterface(Components.interfaces.nsILocalDatebookSyncRecord);
    var remoteRecord = aRemoteRecord.QueryInterface(Components.interfaces.nsIPalmDatebookSyncRecord);
    this.copyEvent(remoteRecord.remoteEvent, localRecord.localEvent);
    remoteRecord.remoteEvent.syncId = remoteRecord.id;
dump("Update local: "+remoteRecord.id+"\n");

    // pretent the alarm has fired when importing an event.
    // Otherwise, the calendar looses its mind and crashes.
    if (localRecord.localEvent.alarm &&
        !localRecord.localEvent.lastAlarmAck &&
        (localRecord.localEvent.start.getTime() < Date.now()))
      localRecord.localEvent.lastAlarmAck = remoteRecord.remoteEvent.start.getTime()

    this.mCalendar.modifyEvent(localRecord.localEvent);
  },

  updateRemote: function (aLocalRecord, aRemoteRecord) {
dump("updateRemote "+aRemoteRecord.id+"  "+aLocalRecord.id+"\n");
    var localRecord = aLocalRecord.QueryInterface(Components.interfaces.nsILocalDatebookSyncRecord);
    var remoteRecord = aRemoteRecord.QueryInterface(Components.interfaces.nsIPalmDatebookSyncRecord);

    this.copyEvent(localRecord.localEvent, remoteRecord.remoteEvent);
    this.mPalmConnection.writeRecord(remoteRecord);

    localRecord.localEvent.syncId = remoteRecord.id;
    this.mCalendar.modifyEvent(localRecord.localEvent);
dump("Tried to update remote "+remoteRecord.description+" to id "+remoteRecord.id+"\n");
  },

  compare: function (aLocalRecord, aRemoteRecord) {
    var localEvent = aLocalRecord.QueryInterface(Components.interfaces.nsILocalDatebookSyncRecord).localEvent;
    var remoteEvent = aRemoteRecord.QueryInterface(Components.interfaces.nsIPalmDatebookSyncRecord).remoteEvent;
dump("compare "+localEvent.title+" to "+remoteEvent.title+"\n");

    if (localEvent.start.getTime() != remoteEvent.start.getTime())
      dump("starttime\n");
    if (localEvent.end.getTime() != remoteEvent.end.getTime())
      dump("endtime\n");
    if (localEvent.title != remoteEvent.title)
      dump("title\n");
    if (localEvent.description != remoteEvent.description)
      dump("description\n");
    if (localEvent.allDay != remoteEvent.allDay)
      dump("allDay\n");
    if (localEvent.alarm != remoteEvent.alarm)
      dump("alarm\n");
    if (localEvent.alarmUnits != remoteEvent.alarmUnits)
      dump("alarmUnits\n");
    if (localEvent.alarmLength != remoteEvent.alarmLength)
      dump("alarmLength\n");
    if (localEvent.recurInterval != remoteEvent.recurInterval)
      dump("recurInterval\n");
    if (localEvent.recurCount != remoteEvent.recurCount)
      dump("recurCount\n");
    if (localEvent.recurEnd.getTime() != remoteEvent.recurEnd.getTime())
      dump("recurEnd.getTime()\n");
    if (localEvent.recur != remoteEvent.recur)
      dump("recur\n");
    if (localEvent.recurUnits != remoteEvent.recurUnits)
      dump("recurUnits\n");
    if (localEvent.recurForever != remoteEvent.recurForever)
      dump("recurForever\n");
    if (localEvent.recurWeekdays != remoteEvent.recurWeekdays)
      dump("recurWeekdays\n");
    if (localEvent.recurWeekNumber != remoteEvent.recurWeekNumber)
      dump("recurWeekNumber\n");

    if ((localEvent.start.getTime() == remoteEvent.start.getTime()) &&
        (localEvent.end.getTime() == remoteEvent.end.getTime()) &&
        (localEvent.title == remoteEvent.title) &&
        (localEvent.description == remoteEvent.description) &&
        (localEvent.allDay == remoteEvent.allDay) &&
        (localEvent.alarm == remoteEvent.alarm) &&
        (localEvent.alarmUnits == remoteEvent.alarmUnits) &&
        (localEvent.alarmLength == remoteEvent.alarmLength) &&
        (localEvent.recurInterval == remoteEvent.recurInterval) &&
        (localEvent.recurCount == remoteEvent.recurCount) &&
        (localEvent.recurEnd.getTime() == remoteEvent.recurEnd.getTime()) &&
        (localEvent.recur == remoteEvent.recur) &&
        (localEvent.recurUnits == remoteEvent.recurUnits) &&
        (localEvent.recurForever == remoteEvent.recurForever) &&
        (localEvent.recurWeekdays == remoteEvent.recurWeekdays) &&
        (localEvent.recurWeekNumber == remoteEvent.recurWeekNumber))
      return true;
     
    return false;
  },

  // Only copy stuff the palm can have changed. We don't want to overwrite info.
  copyEvent: function (aSource, aDest) {
    aDest.start.setTime(aSource.start.getTime());
    aDest.end.setTime(aSource.end.getTime());
    aDest.title = aSource.title;
    aDest.description = aSource.description;
    aDest.allDay = aSource.allDay;
    aDest.alarm = aSource.alarm;
    aDest.alarmUnits = aSource.alarmUnits;
    aDest.alarmLength = aSource.alarmLength;
    aDest.recurInterval = aSource.recurInterval;
    aDest.recurCount = aSource.recurCount;
    aDest.recurEnd.setTime(aSource.recurEnd.getTime());
    aDest.recur = aSource.recur;
    aDest.recurUnits = aSource.recurUnits;
    aDest.recurForever = aSource.recurForever;
    aDest.recurWeekdays = aSource.recurWeekdays;
    aDest.recurWeekNumber = aSource.recurWeekNumber;
  },

  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.nsISyncClient) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
  },
 
  mPalmConnection : null,
  mCalendar : null,
  mFastPossible : false,
  mLastSync : null,
  mLastSyncInUTC : null,
  mCalendarPath : null
}

var syncModule = {
  firstTime: true,

  registerSelf: function (compMgr, fileSpec, location, type) {
    if (this.firstTime) {
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.registerFactoryLocation(this.palmDatebookSyncCID,
                                    "Palm Datebook Sync Component",
                                    this.palmDatebookSyncProgID, 
                                    fileSpec,
                                    location,
                                    type);
  },

  getClassObject: function (compMgr, cid, iid) {
    if (cid.equals(this.palmDatebookSyncCID))
      return this.palmDatebookSyncFactory;

    if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  palmDatebookSyncCID: Components.ID("{01688809-365c-4a12-a622-dcd6041fccda}"),
  palmDatebookSyncProgID: "@mozilla.org/sync/datebook/palm;1",
  palmDatebookSyncFactory: {
    createInstance: function (outer, iid) {
      if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;

      var instance = new nsPalmDatebookSync();
      instance.init();
      return instance.QueryInterface(iid);
    }
  },

  canUnload: function(compMgr) {
    return true;
  }
};

function NSGetModule(compMgr, fileSpec) {
  return syncModule;
}
